ENDC = '\033[0m'


def header(text: str):
    return '\033[95m{}{}'.format(text.rstrip(ENDC), ENDC)


def ok_blue(text: str):
    return '\033[94m{}{}'.format(text.rstrip(ENDC), ENDC)


def ok_cyan(text: str):
    return '\033[96m{}{}'.format(text.rstrip(ENDC), ENDC)


def ok_green(text: str):
    return '\033[92m' + text.rstrip(ENDC) + ENDC


def warn(text: str):
    return '\033[93m' + text.rstrip(ENDC) + ENDC


def fail(text: str):
    return '\033[91m' + text.rstrip(ENDC) + ENDC


def bold(text: str):
    return '\033[1m' + text.rstrip(ENDC) + ENDC


def underline(text: str):
    return '\033[4m' + text.rstrip(ENDC) + ENDC



#!/usr/bin/env python3
import signal
import subprocess
import json
import sys
import os
import logging
import urllib.parse
import urllib.request

import gerrit_tools.config as config
from gerrit_tools import bcolors


cancelled = False


def exec_cmd(cmd: str, warn: bool = True):
    res = None
    try:
        res = subprocess.check_output(cmd, shell=True).decode('UTF-8')
    except subprocess.CalledProcessError as e:
        if warn:
            logging.warning(e)
        return None
    finally:
        return res


def exec_api(method: str, url: str, data: [] = None):
    url = 'https://{}/a/{}'.format(config.GERRIT_URL, url.lstrip('/'))
    if data is not None:
        data = urllib.parse.urlencode(data).encode()
    passwd = urllib.request.HTTPPasswordMgrWithDefaultRealm()
    passwd.add_password(None, url, config.GERRIT_USER, config.GERRIT_API_TOKEN)
    auth_handler = urllib.request.HTTPBasicAuthHandler(passwd)
    opener = urllib.request.build_opener(auth_handler)
    urllib.request.install_opener(opener)
    logging.info('API [{}]: {}'.format(method, url))
    request = urllib.request.Request(url, method=method, data=data)
    with opener.open(request) as f:
        return f.read().decode()


def get_projects(branch: str):
    projects = exec_cmd('{} ls-projects -b {} --format json_compact'.format(config.GERRIT_CMD, branch))
    projects = json.loads(projects)
    has_filter = len(config.manifest_projects) > 0
    if has_filter or len(config.skip_projects) > 0:
        filtered = {}
        for p in projects:
            if cancelled:
                return
            if p in config.skip_projects:
                continue
            if has_filter and p not in config.manifest_projects:
                continue
            filtered[p] = projects[p]
        projects = filtered
    if len(projects) == 0:
        logging.info('No projects found for branch: {}'.format(branch))
        sys.exit(1)
    return projects


def copy_branch(src: str, dst: str, fallback: str = None):
    logging.info('Copy branch: {} => {}'.format(src, dst))
    existing_projects = exec_cmd('{} ls-projects -b {} --format json_compact'.format(config.GERRIT_CMD, dst))
    existing_projects = json.loads(existing_projects).keys()
    projects = get_projects(branch=src)
    for p in projects:
        if cancelled:
            return
        if p in existing_projects:
            print('Branch:', dst, 'already exists in', p)
            continue
        commit = projects[p]['branches'][src]
        print('Creating branch:', dst, 'in', p, 'hash:', commit, end='\t')
        _cmd = '{} create-branch {} {} {}'.format(config.GERRIT_CMD, p, dst, commit)
        if config.noop:
            print('DUMMY: {}'.format(_cmd))
        else:
            create = exec_cmd(_cmd)
            if create is None:
                print('Fail')
            else:
                print('OK')
    if fallback is None:
        return
    fallback_projects = get_projects(branch=fallback)
    for p in fallback_projects:
        if cancelled:
            return
        if p in projects:
            continue
        if p in existing_projects:
            print('Branch', dst, 'already exists in', p)
            continue
        projects[p] = fallback_projects[p]
        commit = projects[p]['branches'][fallback]
        _cmd = '{} create-branch {} {} {}'.format(config.GERRIT_CMD, p, dst, commit)
        if config.noop:
            print('DUMMY: {}'.format(_cmd))
        else:
            print('Creating [fallback] branch:', dst, 'in', p, 'hash:', commit, end='\t')
            create = exec_cmd(_cmd)
            if create is None:
                print('Fail')
            else:
                print('OK')


def delete_branch(branch: str):
    logging.info('Delete branch: {}'.format(branch))
    projects = get_projects(branch=branch)
    for p in projects:
        if cancelled:
            return
        _api_url = '/projects/{}/branches/{}'.format(urllib.parse.quote(p, safe=''),
                                                     urllib.parse.quote(branch, safe=''))
        if config.noop:
            print('DUMMY_API: DELETE {}'.format(_api_url))
        else:
            print('Removing branch:', branch, 'from:', p, end='\t')
            res = exec_api('DELETE', _api_url)
            print('OK')


def list_branch(branch: str):
    logging.info('List branch: {}'.format(branch))
    projects = get_projects(branch=branch)
    if config.verbose:
        from gerrit_tools import bcolors
        for p in projects:
            proj = projects[p]
            links = []
            for link in proj['web_links']:
                links.append('https://{}{}'.format(config.GERRIT_URL, link['url']))
            state = proj['state']
            if state == 'ACTIVE':
                state = bcolors.ok_green(state)
            else:
                state = bcolors.warn(state)
            print(state, proj['branches'][branch], bcolors.bold(bcolors.underline(p)), ' '.join(links))
    else:
        print('\n'.join(projects.keys()))


def repo_upload(branch: str, force: bool, no_thin: bool, unshallow: bool, create: bool):
    config.verbose and print('repo_upload, branch:', branch)
    if len(config.manifest_projects) == 0:
        logging.error('No projects specified in manifest')
        sys.exit(1)

    git_args = '-o skip-validation'
    if force:
        git_args += ' -f'
    if no_thin:
        git_args += ' --no-thin'
    for p in config.manifest_projects:
        if cancelled:
            return
        path = config.manifest_projects[p]
        config.verbose and print(p, 'in', path)
        if not os.path.isdir(path):
            config.verbose and logging.warning('Skip, no path')
            continue

        url = '{}/{}'.format(config.GERRIT_SSH_URL, p)
        print(bcolors.header("New upload for: {}".format(url)))

        if unshallow and not config.noop and os.path.isfile(os.path.join(path, ".git/shallow")):
            with open(os.path.join(path, ".git/shallow")) as f:
                commit = f.readline().strip()
                print(bcolors.warn('Found shallow commit: {}'.format(commit)))
                _tmp_branch = 'gerrit_tools-unshallow'
                exec_cmd('git -C {} checkout --orphan {} {}'.format(path, _tmp_branch, commit))
                exec_cmd('git -C {} commit -C {}'.format(path, commit))
                exec_cmd('git -C {} replace {} {}'.format(path, commit, _tmp_branch))
                exec_cmd('git -C {} filter-branch -- --all'.format(path))

        if create and not config.noop:
            _cmd = '{} create-project {} -p {}'.format(config.GERRIT_CMD, p, config.GERRIT_MANIFEST_REPO)
            if config.noop:
                print('DUMMY: {}'.format(_cmd))
            else:
                print('Create project:', exec_cmd(_cmd) is None and bcolors.warn("Can't") or bcolors.ok_green('Ok'))

        exec_cmd('git -C {} gc'.format(path, config.GERRIT_REMOTE))
        remote = exec_cmd('git -C {} config remote.{}.url'.format(path, config.GERRIT_REMOTE), False)
        if remote is not None:
            remote = remote.strip()
        if remote != url:
            if remote is not None:
                config.verbose and print('redefine remote:', remote, '=>', url)
                _cmd = 'git -C {} remote remove {}'.format(path, config.GERRIT_REMOTE)
                if config.noop:
                    print('DUMMY: {}'.format(_cmd))
                else:
                    exec_cmd(_cmd)
            _cmd = 'git -C {} remote add {} {}'.format(path, config.GERRIT_REMOTE, url)
            if config.noop:
                print('DUMMY: {}'.format(_cmd))
            else:
                exec_cmd(_cmd)
        _cmd = 'git -C {} push {} {} HEAD:refs/heads/{}'.format(path, git_args, config.GERRIT_REMOTE, branch)
        if config.noop:
            print('DUMMY: {}'.format(_cmd))
        else:
            print('Push to branch: {} project: {} in: {}'.format(bcolors.ok_cyan(branch), bcolors.ok_blue(p), path))
            print('\t', exec_cmd(_cmd) is None and bcolors.fail('Fail') or bcolors.ok_green('Ok'))


def main():
    args = config.parse_args()
    if args.command in ['branch', 'b']:
        if args.sub_command in ['copy', 'c']:
            copy_branch(args.source, args.destination, args.fallback_source)
        elif args.sub_command in ['delete', 'd']:
            delete_branch(args.name)
        elif args.sub_command in ['list', 'l']:
            list_branch(args.name)
        else:
            logging.error('Unknown sub-command: {}'.format(args.sub_command))
            sys.exit(1)
    elif args.command in ['repo', 'r']:
        if args.sub_command in ['upload', 'u']:
            repo_upload(args.new_branch_name, args.force, args.no_thin, args.unshallow, args.create)
        else:
            logging.error('Unknown sub-command: {}'.format(args.sub_command))
            sys.exit(1)
    else:
        logging.error('Unknown command: {}'.format(args.command))
        sys.exit(1)


def signal_handler(signum, frame):
    global cancelled
    cancelled = True
    print(bcolors.warn('Cancelled!!!'), signum)
    sys.exit(1)


signal.signal(signal.SIGINT, signal_handler)
if __name__ == "__main__":
    main()



import os
from setuptools import setup, find_packages


def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()


setup(
    name="gerrit_tools",
    version='0.0.8',
    author="Sergey Motyrev",
    author_email="smotyrev@gmail.com",
    description="Gerrit CLI Tool",
    url="https://github.com/smotyrev/gerrit_tools",
    packages=find_packages('.'),
    package_data={'': ['LICENSE', 'README.md']},
    include_package_data=True,
    long_description=read('README.md'),
    zip_safe=False,
    classifiers=[
        'Intended Audience :: Developers',
        'Natural Language :: English',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.5',
        'Environment :: Console',
    ],
    scripts=[],
    platforms=['Any'],
    install_requires=[],
    entry_points={
        "console_scripts": [
            'gerrit_tools = gerrit_tools.main:main',
        ]
    },
)



import argparse
import os
import sys
import logging
from pathlib import Path
from gerrit_tools import bcolors

config: dict = {}
_PKG = 'gerrit_tools'
_DEFAULT_MANIFEST = '.repo/manifest.xml'
_CONFIG_DIR = str(Path.home()) + os.sep + '.config' + os.sep + _PKG
_CONFIG_FILE = _CONFIG_DIR + os.sep + 'config'
if not os.path.isdir(_CONFIG_DIR):
    os.mkdir(_CONFIG_DIR)
if not os.path.isfile(_CONFIG_FILE):
    descriptor = os.open(
        path=_CONFIG_FILE,
        flags=(
                os.O_WRONLY  # access mode: write only
                | os.O_CREAT  # create if not exists
                | os.O_TRUNC  # truncate the file to zero
        ),
        mode=0o600
    )
    with open(descriptor, 'w+'):
        pass
with open(_CONFIG_FILE, 'r+') as rf:
    for line in rf.readlines():
        (k, v) = line.split('\t')
        config[k.strip()] = v.strip()


def save():
    with open(_CONFIG_FILE, 'w') as wf:
        for key in config:
            wf.write('{}\t{}\n'.format(key, config[key]))


def get_val(key: str):
    if key in config:
        return config[key]
    val = input('Enter value for {}: '.format(key))
    config[key] = val
    save()
    return val


GERRIT_URL = get_val('GERRIT_URL')
GERRIT_PORT = get_val('GERRIT_PORT')
GERRIT_USER = get_val('GERRIT_USER')
GERRIT_API_TOKEN = get_val('GERRIT_API_TOKEN')
GERRIT_MANIFEST_REPO = get_val('GERRIT_MANIFEST_REPO')
GERRIT_CMD = 'ssh -p {} -l {} {} gerrit'.format(GERRIT_PORT, GERRIT_USER, GERRIT_URL)
GERRIT_SSH_URL = 'ssh://{}@{}:{}'.format(GERRIT_USER, GERRIT_URL, GERRIT_PORT)
GERRIT_REMOTE = GERRIT_URL.replace('.', '_')

skip_projects = []
manifest: str
manifest_projects = {}
remote_attr: str
noop: bool
verbose: bool
cwd: str


def add_project_list_limiting_args(parser: argparse.ArgumentParser):
    parser.add_argument(
        '-s', '--skip', nargs='+', default=GERRIT_MANIFEST_REPO,
        help='skip projects, default: ' + bcolors.bold(GERRIT_MANIFEST_REPO))
    parser.add_argument(
        '-m', '--manifest', type=str, required=False, default=_DEFAULT_MANIFEST,
        help='use manifest file to limit project list, default: ' + bcolors.bold(_DEFAULT_MANIFEST))
    parser.add_argument(
        '-t', '--manifest-tag', choices=['project', 'remove-project'], default='project',
        help='tag, from where project name obtained, default: project')
    parser.add_argument(
        '-r', '--remote-attr', type=str, required=False,
        help='pick only tags, with provided "remote=..." attribute')
    parser.add_argument(
        '--noop', action='store_true',
        help='don`t execute any modifications to remote server')


def parse_args():
    parser = argparse.ArgumentParser(description='CI repo/git tools.')
    parser.add_argument(
        '-v', '--verbose', action='store_true', help='verbose output')
    parser.add_argument(
        '-V', '--version', action='version')
    parser.add_argument(
        '-C', '--cwd', type=str, default=os.getcwd(), help='change working directory')

    # Sub commands:
    subparsers = parser.add_subparsers(dest='command', help='Sub-commands', required=True)

    ##
    # Branch manipulations
    #
    parser_branch = subparsers.add_parser(
        'branch', aliases='b', help='Manipulate branches')
    parser_branch = parser_branch.add_subparsers(dest='sub_command', help='Sub commands', required=True)
    #   Copy
    parser_branch_cmd = parser_branch.add_parser(
        'copy', aliases='c', help='Copy existing branch to a new one')
    parser_branch_cmd.add_argument(
        'source', type=str, help='Source branch name')
    parser_branch_cmd.add_argument(
        'destination', type=str, help='Destination branch name')
    parser_branch_cmd.add_argument(
        '-f', '--fallback-source', type=str, default=None, help='fallback source branch, if <source> does not exist')
    add_project_list_limiting_args(parser_branch_cmd)
    #   Delete
    parser_branch_cmd = parser_branch.add_parser(
        'delete', aliases='d', help='Delete existing branch')
    parser_branch_cmd.add_argument(
        'name', type=str, help='Branch name')
    add_project_list_limiting_args(parser_branch_cmd)
    #   List
    parser_branch_cmd = parser_branch.add_parser(
        'list', aliases='l', help='List projects, containing branch')
    parser_branch_cmd.add_argument(
        'name', type=str, help='Branch name')
    add_project_list_limiting_args(parser_branch_cmd)
    ##

    ##
    # Repo manipulations
    #
    parser_repo = subparsers.add_parser(
        'repo', aliases='r', help='Bulk manipulation based on deployed repo projects')
    parser_repo = parser_repo.add_subparsers(dest='sub_command', help='Sub commands', required=True)
    #   Upload to branch
    parser_repo_cmd = parser_repo.add_parser(
        'upload', aliases='u',
        help='Upload to new branch, creating remote {} if absent'.format(GERRIT_REMOTE))
    parser_repo_cmd.add_argument(
        '--force', '-f', action='store_true',
        help='Use `git -f` option')
    parser_repo_cmd.add_argument(
        '--create', action='store_true',
        help='Create repo before uploading new branch, parent: {}'.format(bcolors.bold(GERRIT_MANIFEST_REPO)))
    parser_repo_cmd.add_argument(
        '--no-thin', action='store_true',
        help='Use `git push --no-thin` option, in case "Missing commit" errors')
    parser_repo_cmd.add_argument(
        '--unshallow', action='store_true',
        help='If repo has "clone-depth" specified in manifest, use unshallow to trick git.'
             ' Creates new orphan branch, with limited history')
    parser_repo_cmd.add_argument(
        'new_branch_name', type=str,
        help='Branch name, where to upload all repo projects')
    add_project_list_limiting_args(parser_repo_cmd)
    ##

    try:
        args = parser.parse_args()
    except AttributeError:
        import platform
        import pkg_resources
        dist = pkg_resources.get_distribution(_PKG)
        parser.version = '{} (Python {})'.format(bcolors.bold(str(dist)), platform.python_version())
        args = parser.parse_args()

    # change directory
    os.chdir(args.cwd)

    global remote_attr
    remote_attr = args.remote_attr
    global noop
    noop = args.noop
    global verbose
    verbose = args.verbose
    global skip_projects
    skip_projects = args.skip
    global manifest_projects
    if args.manifest is not None:
        tag = args.manifest_tag
        global manifest
        if os.path.isfile(args.manifest):
            manifest = args.manifest
        else:
            manifest = os.getcwd().rstrip(os.sep) + os.sep + args.manifest
        if not os.path.isfile(manifest):
            if args.manifest == _DEFAULT_MANIFEST:
                logging.debug('Manifest', args.manifest, 'not found in', os.getcwd())
                verbose and print(bcolors.warn('Manifest {} not found in {}'.format(args.manifest, os.getcwd())))
            else:
                logging.debug('Manifest', args.manifest, 'not found in', os.getcwd())
        else:
            import xml.etree.ElementTree as ET

            def recursive_read_manifest(m_file: str):
                verbose and print('Read tag <{}> in manifest:'.format(tag), m_file)
                tree = ET.parse(m_file)
                root = tree.getroot()
                if root.tag == 'manifest':
                    for child in root:
                        if child.tag == tag:
                            if remote_attr:
                                if "remote" not in child.attrib or child.attrib["remote"] != remote_attr:
                                    continue
                            name = child.attrib['name']
                            if name not in manifest_projects:
                                if 'path' in child.attrib:
                                    path = child.attrib['path']
                                else:
                                    path = name
                                verbose and print('Append: {} > {}'.format(name, path))
                                manifest_projects[name] = path
                        elif child.tag == 'include':
                            file = os.getcwd().rstrip(os.sep) + os.sep + '.repo/manifests/' + child.attrib['name']
                            recursive_read_manifest(file)
                else:
                    logging.error('Invalid manifest: {}'.format(args.manifest))
                    sys.exit(1)
            recursive_read_manifest(manifest)
            logging.debug('Filter projects: {}'.format(manifest_projects))
    return args
